/*
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package com.inteligr8.wildfly.maven;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;

import org.codehaus.plexus.util.StringUtils;
import org.eclipse.aether.artifact.Artifact;
import org.eclipse.aether.graph.DependencyNode;

import com.inteligr8.wildfly.maven.model.WildflyModule;

/**
 * This class indexes dependencies by their ID (sans version).  It also stores
 * other metadata determined by other executions within this library.
 * 
 * @author brian@inteligr8.com
 */
public class DependencyIndex {
    
    public enum DependencyType {
        SystemModule,
        InSystemModule,
        Module,
        InModule,
        Deployable,
        Self,
        Other
    }
    
    private final Map<String, Artifact> id2node = new HashMap<>();
    private final Map<String, DependencyType> id2type = new HashMap<>();
    private final Map<String, WildflyModule> id2module = new HashMap<>();
    private final Map<String, Boolean> id2accept = new HashMap<>();
    private final Set<String> id2topModule = new HashSet<>();
    
    public String index(DependencyNode node) {
        return this.index(node.getArtifact());
    }
    
    public String index(Artifact artifact) {
        String id = this.toId(artifact);
        if (!this.id2node.containsKey(id))
            this.id2node.put(id, artifact);
        return id;
    }
    
    public void indexAcceptance(String id, boolean accept) {
        this.id2accept.put(id, accept && Boolean.TRUE.equals(this.id2accept.get(id)));
    }
    
    public void indexType(String id, DependencyType dtype) {
        DependencyType existing = this.id2type.get(id);
        if (existing == null) {
            this.id2type.put(id, dtype);
        } else if (dtype.compareTo(existing) < 0) {
            this.id2type.put(id, dtype);
        }
    }
    
    public void indexModule(String id, WildflyModule module) {
        WildflyModule existing = this.id2module.get(id);
        if (existing != null && !existing.equals(module))
            throw new IllegalStateException();
        this.id2module.put(id, module);
    }
    
    public void indexTopLevelModule(String id) {
        this.id2topModule.add(id);
    }
    
    public boolean isIndexed(DependencyNode node) {
        return this.isIndexed(this.toId(node));
    }
    
    public boolean isIndexed(String id) {
        return this.id2node.containsKey(id);
    }
    
    public boolean doAccept(String id) {
        return Boolean.TRUE.equals(this.id2accept.get(id));
    }
    
    public boolean doIgnore(String id) {
        return Boolean.FALSE.equals(this.id2accept.get(id));
    }
    
    public DependencyType getType(String id) {
        return this.id2type.get(id);
    }
    
    public WildflyModule getModule(String id) {
        return this.id2module.get(id);
    }
    
    public boolean isTopLevelModule(String id) {
        return this.id2topModule.contains(id);
    }
    
    public String toId(DependencyNode node) {
        return this.toId(node.getArtifact());
    }
    
    public String toId(Artifact artifact) {
        StringBuilder id = new StringBuilder().append(artifact.getGroupId()).append(':').append(artifact.getArtifactId());
        if (StringUtils.isNotEmpty(artifact.getClassifier()))
            id.append(':').append(artifact.getClassifier());
        id.append(':').append(artifact.getExtension());
        return id.toString();
    }

}
