/*
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package com.inteligr8.wildfly.maven;

import java.util.Collections;
import java.util.List;
import java.util.Stack;

import org.apache.maven.artifact.Artifact;
import org.apache.maven.artifact.DefaultArtifact;
import org.apache.maven.artifact.handler.DefaultArtifactHandler;
import org.apache.maven.artifact.resolver.filter.ExclusionArtifactFilter;
import org.apache.maven.model.Exclusion;
import org.apache.maven.model.building.ModelBuildingException;
import org.apache.maven.model.resolution.UnresolvableModelException;
import org.eclipse.aether.RepositoryException;
import org.eclipse.aether.collection.DependencyGraphTransformationContext;
import org.eclipse.aether.collection.DependencyGraphTransformer;
import org.eclipse.aether.graph.DependencyFilter;
import org.eclipse.aether.graph.DependencyNode;
import org.eclipse.aether.resolution.ArtifactResolutionException;
import org.eclipse.aether.util.artifact.JavaScopes;
import org.eclipse.aether.util.filter.DependencyFilterUtils;
import org.eclipse.aether.util.filter.ScopeDependencyFilter;

import com.inteligr8.wildfly.maven.DependencyIndex.DependencyType;
import com.inteligr8.wildfly.maven.model.WildflyArtifact;
import com.inteligr8.wildfly.maven.model.WildflyDeployable;
import com.inteligr8.wildfly.maven.model.WildflyModule;

public class IndexingDependencyGraphTransformer implements DependencyGraphTransformer {
    
    private final DependencyFilter filter = DependencyFilterUtils.andFilter(
        new ScopeDependencyFilter("test")
    );
    
    private final DependencyAwareComponent dac;
    private final DependencyGraphTransformer transfomer;
    private final DependencyIndex dindex;
    
    public IndexingDependencyGraphTransformer(
            DependencyAwareComponent dac,
            DependencyGraphTransformer transfomer,
            DependencyIndex dindex) {
        this.dac = dac;
        this.transfomer = transfomer;
        this.dindex = dindex;
    }
    
    /**
     * This method is basically an intercepter for the default
     * DependencyGraphTransformer.  It will make the call to perform the
     * default expected functions.  But before that, it will crawl the
     * dependencies to index them.
     */
    @Override
    public DependencyNode transformGraph(DependencyNode node, DependencyGraphTransformationContext context) throws RepositoryException {
        this.crawl(node, new Stack<DependencyNode>(), new Stack<Exclusion>());
        
        return this.transfomer.transformGraph(node, context);
    }
    
    /**
     * This method crawls a dependency and its dependencies, recursively.
     * 
     * @param node A dependency.
     * @param ancestors A stack of ancestral artifacts; parent on top; root on bottom; empty if root.
     */
    private void crawl(DependencyNode node, Stack<DependencyNode> ancestors, Stack<Exclusion> exclusions) {
        if (this.dac.getLog().isDebugEnabled())
            this.dac.getLog().debug("Crawling: " + ancestors + ": " + node);
        if (!new ExclusionArtifactFilter(exclusions).include(this.toMavenArtifact(node.getArtifact()))) {
            if (this.dac.getLog().isDebugEnabled())
                this.dac.getLog().debug("Excluding: " + node);
            return;
        }
        
        if (!this.filter.accept(node, ancestors))
            return;
        
        boolean doExclude = node.getDependency() != null && node.getDependency().getExclusions() != null && !node.getDependency().getExclusions().isEmpty();
        int pushedExclusions = 0;
        if (doExclude) {
            if (this.dac.getLog().isDebugEnabled())
                this.dac.getLog().debug("Will start excluding: " + node.getDependency().getExclusions());
            for (org.eclipse.aether.graph.Exclusion exclusion : node.getDependency().getExclusions()) {
                exclusions.push(this.toMavenExclusion(exclusion));
                pushedExclusions++;
            }
        }
        
        this.index(node, Collections.unmodifiableList(ancestors));
        
        ancestors.push(node);
        for (DependencyNode child : node.getChildren())
            this.crawl(child, ancestors, exclusions);
        ancestors.pop();

        if (doExclude) {
            if (this.dac.getLog().isDebugEnabled())
                this.dac.getLog().debug("Will stop excluding: " + node.getDependency().getExclusions());
            for (int e = 0; e < pushedExclusions; e++)
                exclusions.pop();
        }
    }
    
    /**
     * This method indexes a dependency.
     * 
     * If a dependency encountered more than once, any occurrence inside a
     * system module will take precedence; then a non-system module.
     * 
     * @param node A dependency.
     * @param parents A list of ancestral artifacts; parent is last; root is first; empty if root/self.
     */
    public void index(DependencyNode node, List<DependencyNode> parents) {
        String nodeId = this.dindex.toId(node);
        if (this.dac.getLog().isDebugEnabled())
            this.dac.getLog().debug("Indexing: " + nodeId);
        
        this.dindex.index(node);
        
        if (parents == null || parents.isEmpty()) {
            // root element; the project itself
            // even if it is a module, we aren't going to consider it one in this context
            this.dindex.indexType(nodeId, DependencyType.Self);
            return;
        }
        
        for (DependencyNode ancestor : parents) {
            String ancestorId = this.dindex.toId(ancestor);
            DependencyType type = this.dindex.getType(ancestorId);
            if (type == null)
                throw new IllegalStateException("The dependency parents are expected to be processed before the children");
            
            switch (type) {
                case SystemModule:
                    if (this.dac.getLog().isDebugEnabled())
                        this.dac.getLog().debug(nodeId + ": is in a module");
                    this.dindex.indexType(nodeId, DependencyType.InSystemModule);
                    return;
                case InSystemModule:
                    this.dac.getLog().debug(nodeId + ": is a dependency of something in a module, but not yet indexed (was excluded?)");
                    //this.dindex.indexType(nodeId, DependencyType.InSystemModule);
                    return;
                case Module:
                    if (this.dac.getLog().isDebugEnabled())
                        this.dac.getLog().debug(nodeId + ": is in a module");
                    this.dindex.indexType(nodeId, DependencyType.InModule);
                    return;
                case InModule:
                    this.dac.getLog().debug(nodeId + ": is a dependency of something in a module, but not yet indexed (was excluded?)");
                    return;
                default:
                    // continue
            }
        }
        
        WildflyArtifactExtractor extractor = new WildflyArtifactExtractor(this.dac);
        try {
            WildflyArtifact artifact = extractor.extract(node.getArtifact());
            if (artifact == null) {
                this.dindex.indexType(nodeId, DependencyType.Other);
            } else if (artifact instanceof WildflyDeployable) {
                this.dindex.indexType(nodeId, DependencyType.Deployable);
            } else {
                WildflyModule module = (WildflyModule) artifact;
                module.export = true;
                if (JavaScopes.RUNTIME.equals(node.getDependency().getScope()))
                    module.export = false;
                
                this.dindex.indexType(nodeId, module.isSystem ? DependencyType.SystemModule : DependencyType.Module);
                this.dindex.indexModule(nodeId, module);
                this.dindex.indexTopLevelModule(nodeId);

                if (node.getArtifact().getClassifier().length() > 0) {
                    String shadowedNodeId = node.getArtifact().getGroupId() + ":" + node.getArtifact().getArtifactId() + ":jar";
                    if (this.dac.getLog().isDebugEnabled())
                        this.dac.getLog().debug("Indexing shadowed node: " + shadowedNodeId);
                    this.dindex.indexType(shadowedNodeId, DependencyType.InModule);
                }
            }
        } catch (UnresolvableModelException | ModelBuildingException | ArtifactResolutionException e) {
            this.dac.getLog().warn("Filtering out dependency '" + node.getDependency() + "' due to exception", e);
        }
    }
    
    private Artifact toMavenArtifact(org.eclipse.aether.artifact.Artifact aetherArtifact) {
        Artifact artifact = new DefaultArtifact(aetherArtifact.getGroupId(), aetherArtifact.getArtifactId(), aetherArtifact.getVersion(), DefaultArtifact.SCOPE_COMPILE, aetherArtifact.getExtension(), aetherArtifact.getClassifier(), new DefaultArtifactHandler());
        artifact.setGroupId(aetherArtifact.getGroupId());
        artifact.setArtifactId(aetherArtifact.getArtifactId());
        return artifact;
    }
    
    private Exclusion toMavenExclusion(org.eclipse.aether.graph.Exclusion aetherExclusion) {
        Exclusion exclusion = new Exclusion();
        exclusion.setGroupId(aetherExclusion.getGroupId());
        exclusion.setArtifactId(aetherExclusion.getArtifactId());
        return exclusion;
    }

}
