/*
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package com.inteligr8.wildfly.maven;

import java.util.List;

import org.eclipse.aether.graph.DependencyFilter;
import org.eclipse.aether.graph.DependencyNode;

public class ModuleExclusionDependencyFilter implements DependencyFilter {
    
    public enum Option {
        // this will filter all modules and their children
        FilterAllModules,
        
        // this will filter all children of modules, including modules that are children of other modules
        // but it will include the first module encountered in each part of the tree
        FilterTopLevelModuleChildren,
        
        // this will filter all non-modules and all module children, including modules that are children of other modules
        // this is basically only modules, but those modules could be deep in the tree
        FilterNonModulesAndChildren
    }
    
    private final DependencyAwareComponent dac;
    private final DependencyIndex dindex;
    private final Option option;
    
    public ModuleExclusionDependencyFilter(DependencyAwareComponent dac, DependencyIndex dindex) {
        this(dac, dindex, Option.FilterAllModules);
    }
    
    public ModuleExclusionDependencyFilter(DependencyAwareComponent dac, DependencyIndex dindex, Option option) {
        this.dac = dac;
        this.dindex = dindex;
        this.option = option;
    }
    
    @Override
    public boolean accept(DependencyNode node, List<DependencyNode> parents) {
        String nodeId = this.dindex.toId(node);
        
        if (parents == null || parents.isEmpty()) {
            // root element; the project itself
            this.dindex.indexAcceptance(nodeId, true);
            return true;
        }
        
        if (this.dindex.doAccept(nodeId)) {
            if (this.dac.getLog().isDebugEnabled())
                this.dac.getLog().debug(nodeId + ": already inspected and accepted");
            return true;
        } else if (this.dindex.doIgnore(nodeId)) {
            if (this.dac.getLog().isDebugEnabled())
                this.dac.getLog().debug(nodeId + ": already inspected and ignored");
            return false;
        }
        
        if (this.dindex.getType(nodeId) == null) {
            this.dac.getLog().warn(nodeId + ": missing from index");
            return false;
        }

        switch (this.dindex.getType(nodeId)) {
            case SystemModule:
            case Module:
                if (this.dac.getLog().isDebugEnabled())
                    this.dac.getLog().debug(nodeId + ": is a module");
                switch (this.option) {
                    case FilterTopLevelModuleChildren:
                    case FilterNonModulesAndChildren:
                        return this.dindex.isTopLevelModule(nodeId);
                    default:
                        return false;
                }
            case InSystemModule:
            case InModule:
                if (this.dac.getLog().isDebugEnabled())
                    this.dac.getLog().debug(nodeId + ": is in a module");
                return false;
            default:
                switch (this.option) {
                    case FilterNonModulesAndChildren:
                        return false;
                    default:
                        return true;
                }
        }
    }

}
