/*
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package com.inteligr8.wildfly.maven;

import java.util.HashMap;
import java.util.Map;
import java.util.Properties;

import org.apache.commons.lang3.StringUtils;
import org.apache.maven.model.Model;
import org.apache.maven.model.building.DefaultModelBuildingRequest;
import org.apache.maven.model.building.ModelBuildingException;
import org.apache.maven.model.building.ModelBuildingRequest;
import org.apache.maven.model.building.ModelBuildingResult;
import org.apache.maven.model.building.ModelSource;
import org.apache.maven.model.resolution.UnresolvableModelException;
import org.apache.maven.plugin.logging.Log;
import org.eclipse.aether.artifact.Artifact;
import org.eclipse.aether.artifact.DefaultArtifact;
import org.eclipse.aether.resolution.ArtifactRequest;
import org.eclipse.aether.resolution.ArtifactResolutionException;
import org.eclipse.aether.resolution.ArtifactResult;

import com.inteligr8.wildfly.maven.model.WildflyArtifact;
import com.inteligr8.wildfly.maven.model.WildflyDeployable;
import com.inteligr8.wildfly.maven.model.WildflyModule;

public class WildflyArtifactExtractor {
    
    private final DependencyAwareComponent dac;
    private final DefaultModelBuildingRequest modelBuildRequestTemplate;
    
    public WildflyArtifactExtractor(DependencyAwareComponent dac) {
        this.dac = dac;

        this.modelBuildRequestTemplate = new DefaultModelBuildingRequest()
                .setValidationLevel(ModelBuildingRequest.VALIDATION_LEVEL_MINIMAL)
                .setUserProperties(this.dac.getMavenSession().getUserProperties())
                .setSystemProperties(this.dac.getMavenSession().getSystemProperties())
                .setModelResolver(this.dac.getProjectModelResolver());
    }
    
    public WildflyArtifact extract(Artifact artifact)
            throws UnresolvableModelException, ModelBuildingException, ArtifactResolutionException {
        if (this.getLog().isDebugEnabled())
            this.getLog().debug("Inspecting artifact POM: " + artifact);
        
        if (artifact.getProperty("id", null) != null) {
            String type = artifact.getProperty("type", null);
            if (type != null) {
                WildflyDeployable deployable = new WildflyDeployable();
                deployable.id = artifact.getProperty("id", null);
                deployable.version = artifact.getProperty("version", null);
                deployable.type = type;
                return deployable;
            } else {
                WildflyModule module = new WildflyModule();
                module.id = artifact.getProperty("id", null);
                module.version = artifact.getProperty("version", null);
                module.exportMetaInf = artifact.getProperty("exportMetaInf", null);
                module.exportServices = artifact.getProperty("exportServices", null);
                module.exportAnnotations = Boolean.parseBoolean(artifact.getProperty("exportAnnotations", Boolean.FALSE.toString()));
                module.isSystem = Boolean.parseBoolean(artifact.getProperty("isSystem", Boolean.FALSE.toString()));
                return module;
            }
        }

        DefaultModelBuildingRequest modelBuildRequest = new DefaultModelBuildingRequest(this.modelBuildRequestTemplate);

        // need to go after the artifact as there could be a classifier
        Artifact pomArtifact = new DefaultArtifact(artifact.getGroupId(), artifact.getArtifactId(), artifact.getClassifier(), "pom", artifact.getVersion());
        ArtifactRequest artifactRequest = new ArtifactRequest(pomArtifact, this.dac.getMavenProject().getRemoteProjectRepositories(), null);
        try {
            ArtifactResult artifactResult = this.dac.getArtifactResolver().resolveArtifact(this.dac.getMavenSession().getRepositorySession(), artifactRequest);
            if (this.getLog().isDebugEnabled())
                this.getLog().debug("Found artifact classifier POM [" + artifact + "]: " + artifactResult.getArtifact().getFile());
            modelBuildRequest.setPomFile(artifactResult.getArtifact().getFile());
        } catch (ArtifactResolutionException are) {
            if (this.getLog().isDebugEnabled())
                this.getLog().debug("There is no artifact classifier POM [" + artifact + "]: " + are.getMessage());
            
            ModelSource source = this.dac.getProjectModelResolver().resolveModel(artifact.getGroupId(), artifact.getArtifactId(), artifact.getVersion());
            if (this.getLog().isDebugEnabled())
                this.getLog().debug("Found artifact POM [" + artifact + "]: " + source.getLocation());
            
            modelBuildRequest.setModelResolver(this.dac.getProjectModelResolver());
            modelBuildRequest.setModelSource(source);
        }
        
        ModelBuildingResult result = this.dac.getModelBuilder().build(modelBuildRequest);
        if (this.getLog().isDebugEnabled())
            this.getLog().debug("Parsed artifact POM [" + artifact + "]: " + result.getModelIds());
        
        Model model = result.getEffectiveModel();
        if (this.getLog().isDebugEnabled())
            this.getLog().debug("Formulated artifact effective POM [" + artifact + "]: " + model.getId());
        
        if (!"client".equals(artifact.getClassifier())) {
            switch (model.getPackaging().toLowerCase()) {
                case "ear":
                case "war":
                case "ejb":
                    this.getLog().debug("Discovered deployable artifact [" + artifact + "]: " + model.getId());
    
                    WildflyDeployable deployable = new WildflyDeployable();
                    deployable.id = artifact.getGroupId() + ":" + artifact.getArtifactId() + ":" + artifact.getClassifier();
                    deployable.version = artifact.getVersion();
                    deployable.type = model.getPackaging();
                    
                    Map<String, String> props = new HashMap<>();
                    props.put("id", deployable.id);
                    props.put("version", deployable.version);
                    props.put("type", deployable.type);
                    artifact.setProperties(props);
                    
                    return deployable;
                default:
            }
        }

        Properties modelProps = model.getProperties();
        String moduleId = modelProps.getProperty("wildfly.module.id");
        if (moduleId == null) {
            if (this.getLog().isDebugEnabled())
                this.getLog().debug("No 'wildfly.module.id' property; not a Wildfly module project; adding as module resource: " + artifact);
            return null;
        } else {
            this.getLog().debug("Discovered module artifact [" + artifact + "]: " + model.getId());
            String moduleVersion = StringUtils.trimToNull(modelProps.getProperty("wildfly.module.version"));
            
            WildflyModule module = new WildflyModule();
            module.id = moduleId;
            module.version = moduleVersion;
            module.exportMetaInf = modelProps.getProperty("wildfly.module.meta");
            module.exportServices = modelProps.getProperty("wildfly.module.services");
            module.exportAnnotations = Boolean.parseBoolean(modelProps.getProperty("wildfly.module.annotations"));
            module.isSystem = Boolean.parseBoolean(modelProps.getProperty("wildfly.module.system"));
            
            // for backward compatibility
            if ("true".equalsIgnoreCase(module.exportMetaInf))
                module.exportMetaInf = "export";
            
            Map<String, String> props = new HashMap<>();
            props.put("id", module.id);
            props.put("version", module.version);
            props.put("exportMetaInf", String.valueOf(module.exportMetaInf));
            props.put("exportServices", module.exportServices);
            props.put("exportAnnotations", String.valueOf(module.exportAnnotations));
            props.put("isSystem", String.valueOf(module.isSystem));
            artifact.setProperties(props);
            
            return module;
        }
    }
    
    private Log getLog() {
        return this.dac.getLog();
    }

}
