/*
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package com.inteligr8.wildfly.maven.goal;

import java.io.File;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashSet;
import java.util.Set;

import javax.inject.Inject;

import org.apache.maven.execution.MavenSession;
import org.apache.maven.model.building.ModelBuilder;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.plugins.annotations.Parameter;
import org.apache.maven.project.DefaultDependencyResolutionRequest;
import org.apache.maven.project.DependencyResolutionException;
import org.apache.maven.project.DependencyResolutionRequest;
import org.apache.maven.project.DependencyResolutionResult;
import org.apache.maven.project.MavenProject;
import org.apache.maven.project.ProjectDependenciesResolver;
import org.apache.maven.project.ProjectModelResolver;
import org.codehaus.plexus.component.annotations.Requirement;
import org.eclipse.aether.DefaultRepositorySystemSession;
import org.eclipse.aether.RepositorySystem;
import org.eclipse.aether.RepositorySystemSession;
import org.eclipse.aether.collection.DependencyGraphTransformer;
import org.eclipse.aether.graph.Dependency;
import org.eclipse.aether.graph.DependencyFilter;
import org.eclipse.aether.impl.ArtifactResolver;
import org.eclipse.aether.impl.RemoteRepositoryManager;
import org.eclipse.aether.util.artifact.JavaScopes;
import org.eclipse.aether.util.filter.DependencyFilterUtils;
import org.eclipse.aether.util.filter.ScopeDependencyFilter;

import com.inteligr8.wildfly.maven.DependencyAwareComponent;
import com.inteligr8.wildfly.maven.DependencyIndex;
import com.inteligr8.wildfly.maven.IndexingDependencyGraphTransformer;
import com.inteligr8.wildfly.maven.ModuleExclusionDependencyFilter;
import com.inteligr8.wildfly.maven.ModuleExclusionDependencyFilter.Option;
import com.inteligr8.wildfly.maven.ModuleProvidedDependencyFilter;
import com.inteligr8.wildfly.maven.ValidationDependencyFilter;
import com.inteligr8.wildfly.maven.model.WildflyModule;
import com.inteligr8.wildfly.maven.util.DirectoryAccessor;

/**
 * @author brian@inteligr8.com
 */
public abstract class AbstractGoal extends DisablableGoal implements DependencyAwareComponent {

    @Parameter( defaultValue = "${session}", readonly = true )
    protected MavenSession session;
    
    @Parameter( defaultValue = "${project}", readonly = true )
    protected MavenProject project;

    @Parameter( property = "wildfly.build.outputDirectory", required = true, defaultValue = "${project.build.directory}/wildfly" )
    protected File outputDirectory;
    
    @Requirement
    private RemoteRepositoryManager rrm;
    
    @Requirement
    private ArtifactResolver artifactResolver;
    
    @Requirement
    private ModelBuilder pomBuilder;
    
    @Requirement
    private RepositorySystem repositorySystem;
    
    @Requirement
    private ProjectDependenciesResolver depResolver;

    @Inject
    protected DirectoryAccessor directoryAccessor;
    
    protected ProjectModelResolver pomResolver;
	protected Set<String> excludeScopes = new HashSet<String>(Arrays.asList(JavaScopes.PROVIDED, JavaScopes.TEST));
	
	@Override
	public final void executeEnabled() throws MojoExecutionException, MojoFailureException {
	    if (this.getLog().isDebugEnabled()) {
            this.getLog().debug("Repository System: " + this.repositorySystem);
    	    this.getLog().debug("Remote Repository Manager: " + this.rrm);
            this.getLog().debug("Local Repository: " + this.session.getLocalRepository());
            this.getLog().debug("Repositories: " + this.project.getRemoteProjectRepositories());
	    }
	    
	    this.pomResolver = new ProjectModelResolver(
	            this.session.getRepositorySession(),
	            null,
	            this.repositorySystem,
	            this.rrm,
	            this.project.getRemoteProjectRepositories(),
	            null,
	            null);
	    
	    this.executePre();

	    DependencyIndex dindex = new DependencyIndex();
        RepositorySystemSession session = this.injectSession(dindex);
	    
        this.crawlDependencies(session, dindex);
        this.handleDependencies(dindex);
        
        this.executePost();
    }
	
	/**
	 * This method creates a Maven RepositorySystemSession that is able to load
	 * our DependencyIndex.
	 */
	private RepositorySystemSession injectSession(DependencyIndex dindex) {
        DefaultRepositorySystemSession session = new DefaultRepositorySystemSession(this.session.getRepositorySession());
        DependencyGraphTransformer transformer = session.getDependencyGraphTransformer();
	    transformer = new IndexingDependencyGraphTransformer(this, transformer, dindex);
        session.setDependencyGraphTransformer(transformer);
        return session;
	}
    
    private DependencyResolutionResult crawlDependencies(
            RepositorySystemSession session,
            DependencyIndex dindex)
                    throws MojoFailureException, MojoExecutionException {
        return this.crawlDependencies(session, Arrays.asList(
                new ScopeDependencyFilter("test"),
                new ValidationDependencyFilter(this, dindex),
                new ModuleProvidedDependencyFilter(this, dindex)));
	}
	
	private DependencyResolutionResult crawlDependencies(
	        RepositorySystemSession session,
	        Collection<DependencyFilter> filters)
	                throws MojoFailureException, MojoExecutionException {
        DependencyResolutionRequest request = new DefaultDependencyResolutionRequest()
                .setMavenProject(this.project)
                .setRepositorySession(session)
                .setResolutionFilter(DependencyFilterUtils.andFilter(filters));
        
        // this will resolve all dependencies, recursively
        // any filtered out dependency will filter out its recursive dependencies too
        try {
            if (this.getLog().isDebugEnabled())
                this.getLog().debug("Finding dependency artifacts ...");
            return this.depResolver.resolve(request);
        } catch (DependencyResolutionException dre) {
            throw new MojoExecutionException("Dependency resolution failed", dre);
        }
	}
    
    private void handleDependencies(DependencyIndex dindex) throws MojoFailureException, MojoExecutionException {
        DependencyResolutionResult result = this.crawlDependencies(
                this.session.getRepositorySession(),
                this.getDependencyFilters(dindex));
        
        if (!result.getUnresolvedDependencies().isEmpty())
            throw new MojoFailureException("Some dependencies could not be resolved: " + result.getUnresolvedDependencies());
        
        for (Dependency dependency : result.getResolvedDependencies()) {
            if (this.getLog().isDebugEnabled())
                this.getLog().debug("Found dependency artifact: " + dependency.getArtifact());
            
            String id = dindex.toId(dependency.getArtifact());
            switch (dindex.getType(id)) {
                case SystemModule:
                case Module:
                    this.handleModuleDependency(dependency, dindex.getModule(id));
                    break;
                case Deployable:
                    this.handleDeployableDependency(dependency);
                    break;
                default:
                    if (!"pom".equals(dependency.getArtifact().getExtension()))
                        this.handleDependency(dependency);
            }
        }
    }
	
	protected Collection<DependencyFilter> getDependencyFilters(DependencyIndex dindex) {
        return Arrays.asList(
                new ScopeDependencyFilter("test", "provided"),
                // this filter will remove all Wildfly Module dependencies, recursively
                new ModuleExclusionDependencyFilter(this, dindex, Option.FilterTopLevelModuleChildren));
    }
    
    protected abstract void executePre() throws MojoExecutionException, MojoFailureException;
	
	protected abstract void handleDependency(Dependency dependency) throws MojoExecutionException, MojoFailureException;
	
	protected void handleDeployableDependency(Dependency dependency) throws MojoExecutionException, MojoFailureException {
    }
    
    protected abstract void handleModuleDependency(Dependency dependency, WildflyModule moduleId) throws MojoExecutionException, MojoFailureException;

    protected abstract void executePost() throws MojoExecutionException, MojoFailureException;
    
    @Override
    public MavenSession getMavenSession() {
        return this.session;
    }
    
    @Override
    public MavenProject getMavenProject() {
        return this.project;
    }
    
    @Override
    public ProjectModelResolver getProjectModelResolver() {
        return this.pomResolver;
    }
    
    @Override
    public ModelBuilder getModelBuilder() {
        return this.pomBuilder;
    }

    @Override
    public ArtifactResolver getArtifactResolver() {
        return artifactResolver;
    }

}
