/*
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package com.inteligr8.wildfly.maven.goal;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.nio.file.FileVisitResult;
import java.nio.file.FileVisitor;
import java.nio.file.Files;
import java.nio.file.Path;
import java.nio.file.attribute.BasicFileAttributes;
import java.util.ArrayList;
import java.util.Collection;
import java.util.List;
import java.util.zip.ZipEntry;
import java.util.zip.ZipOutputStream;

import javax.inject.Inject;

import org.apache.maven.execution.MavenSession;
import org.apache.maven.model.Model;
import org.apache.maven.model.Profile;
import org.apache.maven.model.building.DefaultModelBuildingRequest;
import org.apache.maven.model.building.ModelBuilder;
import org.apache.maven.model.building.ModelBuildingException;
import org.apache.maven.model.building.ModelBuildingRequest;
import org.apache.maven.model.building.ModelBuildingResult;
import org.apache.maven.model.io.ModelWriter;
import org.apache.maven.plugin.AbstractMojo;
import org.apache.maven.plugin.MojoExecutionException;
import org.apache.maven.plugin.MojoFailureException;
import org.apache.maven.plugins.annotations.LifecyclePhase;
import org.apache.maven.plugins.annotations.Mojo;
import org.apache.maven.plugins.annotations.Parameter;
import org.apache.maven.project.MavenProject;
import org.apache.maven.project.MavenProjectHelper;
import org.codehaus.plexus.component.annotations.Component;
import org.codehaus.plexus.util.FileUtils;

import com.inteligr8.wildfly.maven.util.DirectoryAccessor;

/**
 * @author brian@inteligr8.com
 */
@Mojo( name = "module-package", threadSafe = false, defaultPhase = LifecyclePhase.PACKAGE )
@Component( role = org.apache.maven.plugin.Mojo.class )
public class PackageModuleGoal extends AbstractMojo {

    @Parameter( defaultValue = "${session}", readonly = true )
    protected MavenSession session;
    
    @Parameter( defaultValue = "${project}", readonly = true )
    protected MavenProject project;

    @Parameter( property = "wildfly.module.classifier", required = false )
    protected String moduleClassifier;

    @Parameter( property = "wildfly.module.id" )
    protected String moduleId;

    @Parameter( property = "wildfly.module.version", defaultValue = "main" )
    protected String moduleVersion;

    @Parameter( property = "wildfly.build.outputDirectory", required = true, defaultValue = "${project.build.directory}/wildfly" )
    protected File packageDirectory;

    @Inject
    protected DirectoryAccessor directoryAccessor;

    @Inject
    protected MavenProjectHelper helper;

    @Inject
    protected ModelBuilder modelBuilder;

    @Inject
    protected ModelWriter modelWriter;
    
    @Override
    public void execute() throws MojoExecutionException, MojoFailureException {
        boolean isModuleOnly = "pom".equals(this.project.getPackaging());

        File buildDirectory = this.directoryAccessor.validateAndInitializeDirectory(new File(this.project.getBuild().getDirectory()));
        if (!isModuleOnly) {
            if (this.moduleId == null)
                throw new MojoExecutionException("A 'moduleId' configuration or 'wildfly.module.id' property is required");
            this.addThisArtifactToModule(buildDirectory);
        }

        if (!isModuleOnly) {
            File zipFile = this.createZip(buildDirectory, this.moduleClassifier, "zip");
            this.helper.attachArtifact(project, "zip", this.moduleClassifier, zipFile);

            File file = this.createModulePom(buildDirectory);
            this.helper.attachArtifact(project, "pom", this.moduleClassifier, file);
        } else {
            this.createZip(buildDirectory, "zip");
        }
    }

    private void addThisArtifactToModule(File buildDirectory) throws MojoExecutionException {
        File packageDirectory = this.directoryAccessor.validateAndInitializeDirectory(this.packageDirectory, this.moduleId, this.moduleVersion);
        File file = new File(buildDirectory, this.project.getBuild().getFinalName() + "." + this.project.getPackaging());

        try {
            FileUtils.copyFileToDirectory(file, packageDirectory);
        } catch (IOException ie) {
            throw new MojoExecutionException("An I/O related issue occurred", ie);
        }

        if (this.getLog().isInfoEnabled())
            this.getLog().info("Copied artifact: " + this.project.getArtifact().getId());
    }

    private File createZip(File buildDirectory, String ext) throws MojoExecutionException {
        File zipfile = new File(buildDirectory, this.project.getBuild().getFinalName() + "." + ext);
        this.createZip(buildDirectory, zipfile);
        return zipfile;
    }

    private File createZip(File buildDirectory, String classifier, String ext) throws MojoExecutionException {
        File zipfile = new File(buildDirectory, this.project.getBuild().getFinalName() + "-" + classifier + "." + ext);
        this.createZip(buildDirectory, zipfile);
        return zipfile;
    }

    private void createZip(File buildDirectory, File zipfile) throws MojoExecutionException {
        try {
            FileOutputStream fostream = new FileOutputStream(zipfile);
            try {
                BufferedOutputStream bostream = new BufferedOutputStream(fostream);
                ZipOutputStream zostream = new ZipOutputStream(bostream);
                try {
                    this.zipModule(zostream);
                } finally {
                    zostream.close();
                }
            } finally {
                fostream.close();
            }
        } catch (IOException ie) {
            throw new MojoExecutionException("An I/O issue occurred", ie);
        }
    }
    
    private void zipModule(ZipOutputStream zostream) throws IOException {
        Path path = this.packageDirectory.toPath();
        if (getLog().isDebugEnabled())
            getLog().debug("Zipping files in: " + path);
        
        Files.walkFileTree(path, new FileVisitor<Path>() {
            @Override
            public FileVisitResult preVisitDirectory(Path dir, BasicFileAttributes attrs) {
                return FileVisitResult.CONTINUE;
            }
            
            @Override
            public FileVisitResult postVisitDirectory(Path dir, IOException exc) throws IOException {
                if (exc == null) return FileVisitResult.CONTINUE;
                throw exc;
            }
            
            @Override
            public FileVisitResult visitFile(Path file, BasicFileAttributes attrs) throws IOException {
                Path zippath = path.relativize(file);
                if (getLog().isDebugEnabled())
                    getLog().debug("Zipping file: " + zippath);
                
                ZipEntry zentry = new ZipEntry(zippath.toString().replace('\\', '/'));
                zostream.putNextEntry(zentry);
                try {
                    Files.copy(file, zostream);
                } finally {
                    zostream.closeEntry();
                }
                
                return FileVisitResult.CONTINUE;
            }
            
            @Override
            public FileVisitResult visitFileFailed(Path file, IOException exc) throws IOException {
                throw exc;
            }
        });
    }

    private File createModulePom(File buildDirectory) throws MojoExecutionException {
        ModelBuildingRequest request = new DefaultModelBuildingRequest()
                .setPomFile(this.project.getFile())
                .setActiveProfileIds(this.getProfileIds(this.project.getActiveProfiles()));
        try {
            ModelBuildingResult result = this.modelBuilder.build(request);
            Model model = result.getEffectiveModel();
            model.setPackaging("pom");

            File modulePomFile = new File(buildDirectory, this.project.getBuild().getFinalName() + "-" + moduleClassifier + ".xml");
            this.modelWriter.write(modulePomFile, null, model);
            return modulePomFile;
        } catch (ModelBuildingException mbe) {
            throw new MojoExecutionException("The module POM failed to generate", mbe);
        } catch (IOException io) {
            throw new MojoExecutionException("An I/O issue occurred", io);
        }
    }

    private List<String> getProfileIds(Collection<Profile> profiles) {
        List<String> ids = new ArrayList<>(profiles.size());
        for (Profile profile : profiles)
            ids.add(profile.getId());
        return ids;
    }

}
