/*
 * This program is free software: you can redistribute it and/or modify it
 * under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or (at your
 * option) any later version.
 * 
 * This program is distributed in the hope that it will be useful, but WITHOUT
 * ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License for
 * more details.
 * 
 * You should have received a copy of the GNU General Public License along
 * with this program.  If not, see <https://www.gnu.org/licenses/>.
 */
package com.inteligr8.wildfly.maven.xml;

import java.io.BufferedOutputStream;
import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.OutputKeys;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Element;

import com.inteligr8.wildfly.maven.model.WildflyModule;

public abstract class AbstractXmlBuilder {

    protected final int chunkSize = 16 * 1024;
    protected final TransformerFactory tfactory = TransformerFactory.newInstance();
    protected final DocumentBuilderFactory dbfactory = DocumentBuilderFactory.newInstance();
    protected final String ns;
    protected final Document xmldoc;
    protected final Element dependenciesElement;
    protected final boolean enableSlotAttribute;
    
    public AbstractXmlBuilder(String ns, boolean enableSlotAttribute) throws ParserConfigurationException {
        DocumentBuilder docbuilder = this.dbfactory.newDocumentBuilder();
        
        this.ns = ns;
        this.xmldoc = docbuilder.newDocument();
        this.dependenciesElement = this.xmldoc.createElementNS(this.ns, "dependencies");
        this.enableSlotAttribute = enableSlotAttribute;
    }
    
    public abstract String getXmlFilename();
    
    public void addDependency(WildflyModule module, boolean isDeployment) {
        String name = module.id;
        if (!this.enableSlotAttribute && module.version != null && !module.version.equals("main"))
            name += ":" + module.version;
        
        Element moduleElement = this.xmldoc.createElementNS(this.ns, "module");
        moduleElement.setAttribute("name", name);
        if (this.enableSlotAttribute && module.version != null && !module.version.equals("main"))
            moduleElement.setAttribute("slot", module.version);
        if (module.export)
            moduleElement.setAttribute("export", Boolean.TRUE.toString());
        if (isDeployment && module.exportMetaInf != null)
            moduleElement.setAttribute("meta-inf", module.exportMetaInf);
        if (module.exportServices != null)
            moduleElement.setAttribute("services", module.exportServices);
        if (isDeployment && module.exportAnnotations)
            moduleElement.setAttribute("annotations", Boolean.TRUE.toString());
        this.dependenciesElement.appendChild(moduleElement);
    }
    
    public void writeTo(File directory) throws IOException, TransformerException {
        Transformer transformer = this.tfactory.newTransformer();
        transformer.setOutputProperty(OutputKeys.INDENT, Boolean.TRUE.toString());
        transformer.setOutputProperty(OutputKeys.INDENT, "yes");
        
        File moduleFile = new File(directory, this.getXmlFilename());
        FileOutputStream fostream = new FileOutputStream(moduleFile, false);
        try {
            BufferedOutputStream bostream = new BufferedOutputStream(fostream, this.chunkSize);
            transformer.transform(new DOMSource(this.xmldoc), new StreamResult(bostream));
        } finally {
            fostream.close();
        }
    }

}
